"""
getting LuminanceTest trials table (for testing pupil size) requires:
- one or more hdf5 files with trial/stim data, called 'Luminance_*.hdf5'
- frame display times from sync, to use in place of frametimes in hdf5 file
- latency estimate for each stim presentation, to be added to frame
display times to get stim onset times

# >>> stim = "//allen/programs/mindscope/workgroups/dynamicrouting/PilotEphys/Task 2 pilot/DRpilot_708019_20240322/LuminanceTest_708019_20240322_153324.hdf5"
# >>> sync = npc_sync.get_sync_data('//allen/programs/mindscope/workgroups/dynamicrouting/PilotEphys/Task 2 pilot/DRpilot_708019_20240322/20240322T153301.h5')
# >>> lum_trials = LuminanceTest(stim, sync)
# >>> assert not lum_trials.to_dataframe().empty
"""

from __future__ import annotations

from collections.abc import Iterable

import npc_io
import npc_stim
import npc_sync
import numpy as np
import numpy.typing as npt

from npc_sessions.trials.TaskControl import TaskControl


class LuminanceTest(TaskControl):

    def __init__(
        self,
        hdf5: npc_io.PathLike,
        sync: npc_sync.SyncPathOrDataset,
        ephys_recording_dirs: Iterable[npc_io.PathLike] | None = None,
        **kwargs,
    ) -> None:
        if sync is None:
            raise ValueError(
                f"sync data is required for {self.__class__.__name__} trials table"
            )
        self._ephys_recording_dirs = ephys_recording_dirs
        super().__init__(
            hdf5, sync, ephys_recording_dirs=ephys_recording_dirs, **kwargs
        )

    def find(self, key: str) -> npt.NDArray[np.bool_] | None:
        if key in self._hdf5_data:
            return ~np.isnan(self._hdf5_data[key][()])
        return None

    @npc_io.cached_property
    def _len_all_trials(self) -> int:
        return len(self._hdf5_data["trialStartFrame"][()])

    @npc_io.cached_property
    def _idx(self) -> npt.NDArray[np.int32]:
        """Used for extracting a subset of inds throughout all properties.
        Must be constructed from private properties directly from the hdf5 file"""
        return np.arange(self._len_all_trials)

    @npc_io.cached_property
    def start_time(self) -> npt.NDArray[np.float64]:
        """falling edge of first vsync in each trial"""
        return npc_stim.safe_index(
            self._flip_times, self._hdf5_data["trialStartFrame"][self._idx]
        )

    @npc_io.cached_property
    def stop_time(self) -> npt.NDArray[np.float64]:
        """falling edge of vsync after stimulus end + inter-stim frames"""
        return npc_stim.safe_index(
            self._flip_times,
            self._hdf5_data["trialStartFrame"][self._idx]
            + self._hdf5_data["framesPerLevel"][()],
        )

    @npc_io.cached_property
    def trial_index(self) -> npt.NDArray[np.int32]:
        return np.arange(self._len_all_trials)[self._idx]

    @npc_io.cached_property
    def level(self) -> npt.NDArray[np.int32]:
        # round because some values end up as -0.40000000000000013
        return np.round(self._hdf5_data["trialLevel"][self._idx], 3)

    @npc_io.cached_property
    def _len(self) -> int:
        return len(self._idx)


if __name__ == "__main__":
    import doctest

    import dotenv

    dotenv.load_dotenv(dotenv.find_dotenv(usecwd=True))
    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )



from __future__ import annotations

import typing
from collections.abc import Iterable, Mapping, Sequence
from typing import Literal, TypeAlias

import hdmf.common
import numpy as np
import numpy.typing as npt
import pandas as pd
import rich.progress

from npc_sessions.utils.misc import get_taskcontrol_intervals_table_name

Interval: TypeAlias = (
    Sequence[float]
    | Mapping[Literal["start_time", "stop_time"], float]
    | pd.Series
    | pd.DataFrame
)
UnitSelection: TypeAlias = int | str | Iterable[int] | pd.DataFrame | pd.Series


def parse_intervals(
    intervals: Interval | Iterable[Interval],
) -> tuple[tuple[float, float], ...]:
    intervals = tuple(intervals)
    is_single_interval = len(intervals) == 2 and isinstance(intervals[0], (int, float))
    intervals = [intervals] if is_single_interval else intervals

    _intervals: list[tuple[float, float]] = []
    for i in intervals:
        times: tuple[float, float]
        try:
            times = (i.get("start_time"), i.get("stop_time"))  # type: ignore [union-attr]
        except AttributeError:
            times = tuple(i)
        if all(isinstance(t, hdmf.common.table.VectorData) for t in times):
            t = typing.cast(
                tuple[hdmf.common.table.VectorData, hdmf.common.table.VectorData], times
            )
            times = (t[0].data, t[1].data)
        if all(isinstance(t, Iterable) for t in times):
            t = typing.cast(tuple[Iterable[float], Iterable[float]], times)
            _intervals.extend(parse_intervals(zip(*t, strict=False)))
            continue
        if None in times or len(times) != 2 or sorted(times) != list(times):
            raise ValueError(
                f"Invalid interval, expected [start_time, stop_time]: {times!r}"
            )
        else:
            _intervals.append(times)
    return tuple(_intervals)


def is_overlap_in_intervals(*intervals: Interval) -> bool:
    """For two or more intervals, check if any overlap.

    >>> is_overlap_in_intervals((0, 1), (1, 2))
    False
    >>> is_overlap_in_intervals((0, 1), (1, 2), (2, 3))
    False
    >>> is_overlap_in_intervals(dict(start_time=0, stop_time=1), dict(start_time=1, stop_time=2))
    False
    >>> is_overlap_in_intervals((0, 1), (0.5, 2))
    True
    >>> is_overlap_in_intervals((0, 1), (0.5, 0.6))
    True
    >>> is_overlap_in_intervals((0, 1), (0.5, 1.5))
    True
    >>> is_overlap_in_intervals((0, 1), (-0.5, 0.5))
    True
    """
    _intervals = list(parse_intervals(intervals))
    if len(_intervals) < 2:
        raise ValueError(f"Expected two or more intervals, got {_intervals!r}")
    for idx in range(len(_intervals) - 1):
        i = _intervals.pop(idx)
        is_overlap = any(
            (i[0] < other[1] and other[0] < i[1])
            or (other[0] < i[1] and i[0] < other[1])
            for other in _intervals
        )
        if is_overlap:
            return True
    return False


def is_within_intervals(interval: Interval, *other_intervals: Interval) -> bool:
    """Check if inteval is completely within any of the other intervals.

    >>> is_within_intervals((0, 1), (1, 2))
    False
    >>> is_within_intervals((0, 1), (1, 2), (2, 3))
    False
    >>> is_within_intervals((0, 1), (0, 2))
    True
    >>> is_within_intervals((0, 1), (0, 2), (2, 3))
    True
    >>> is_within_intervals((0, 1), (0, 0.5))
    False
    >>> is_within_intervals((0, 1), (-0.5, 0.5))
    False
    """
    _intervals = list(parse_intervals([interval, *other_intervals]))
    interval = _intervals.pop(0)
    for other in _intervals:
        is_within = other[0] <= interval[0] and interval[1] <= other[1]
        if is_within:
            return True
    return False


def is_valid_interval(session, interval: Interval) -> bool:
    """Check if time interval is valid, based on `invalid_times`"""
    if session.invalid_times is not None:
        if any(
            is_overlap_in_intervals(interval, invalid_interval)
            for invalid_interval in parse_intervals(session.invalid_times)
        ):
            return False
    return True


def parse_units(session, unit_selection: UnitSelection = None) -> pd.DataFrame:
    units: pd.DataFrame
    if hasattr(unit_selection, "spike_times"):
        return unit_selection
    if isinstance(unit_selection, (pd.DataFrame, pd.Series)):
        units = session.units[:].query(f"unit_id in {unit_selection.unit_id.array!r}")
    elif isinstance(unit_selection, str):
        units = session.units[:].query(f"unit_id == {unit_selection!r}")
    elif isinstance(
        unit_selection, (int, Iterable)
    ):  # beware order: str/DataFrame are Iterable
        units = session.units[unit_selection]
    elif unit_selection is None:
        units = session.units[:]
    else:
        raise TypeError(
            f"Expected unit_selection to be {UnitSelection=} - got {unit_selection!r}"
        )
    if unit_selection is not None and len(units) == 0:
        raise ValueError(f"No units found for unit_selection: {unit_selection!r}")
    return units


import numba


@numba.njit(nogil=True)
def get_unit_spike_times(
    unit_spike_times: npt.NDArray[np.floating],
    start_time: float,
    stop_time: float,
) -> npt.NDArray[np.floating]:
    """Get spike times within interval"""
    return unit_spike_times[
        # bisect.bisect_left(unit_spike_times, start_time): bisect.bisect_right(unit_spike_times, stop_time)
        (unit_spike_times >= start_time)
        & (unit_spike_times <= stop_time)
    ]


@numba.njit(nogil=True, parallel=True, fastmath=True)
def get_spike_counts(
    spike_times: npt.NDArray[np.floating],
    intervals: npt.NDArray[np.floating],
) -> npt.NDArray[np.intp]:
    counts = np.diff(
        np.searchsorted(spike_times, intervals)
    )  # axis not supported in np.diff for numba
    return counts


def apply_invalid_intervals(
    session,
    intervals: Interval | Iterable[Interval],
    units_by_intervals: npt.NDArray | Iterable[npt.NDArray],
    unit_selection: UnitSelection | None = None,
) -> npt.NDArray[np.float64]:
    _intervals = parse_intervals(intervals)
    units = parse_units(session, unit_selection)

    # convert to float64 to allow NaNs
    units_by_intervals = np.array(units_by_intervals, dtype=np.float64, copy=True)
    assert units_by_intervals.shape[0:2] == (len(units), len(_intervals))

    is_global_obs_interval = all(
        i == units.iloc[0].obs_intervals for i in units.obs_intervals.array
    )
    # deal with unobserved intervals on a per-unit basis
    for idx, unit in enumerate(units):
        if not is_global_obs_interval:
            units_by_intervals[idx, ...] = np.where(
                [is_within_intervals(i, *unit.obs_intervals) for i in _intervals],
                units_by_intervals[idx, ...],
                np.nan,
            )
    # deal with global unobserved intervals (ie same for all units) and global invalid
    # intervals
    for idx, interval in enumerate(_intervals):
        if not is_valid_interval(session, interval) or (
            is_global_obs_interval
            and not is_within_intervals(interval, *units.iloc[0].obs_intervals)
        ):
            units_by_intervals[:, idx] = np.nan
    return units_by_intervals


def get_spike_counts_in_intervals(
    session,
    intervals: Interval | Iterable[Interval],
    unit_selection: UnitSelection | None = None,
    as_spikes_per_second: bool = False,
) -> npt.NDArray[np.floating]:
    """Get spike counts in interval(s) for unit(s).

    - returns [units x intervals] array of spike counts (as floats)
    - if no spiking data is available within an interval for a unit, its
      count is `np.nan`
    """
    _intervals = parse_intervals(intervals)
    units = parse_units(session, unit_selection)

    _spikes_per_interval_per_unit: list[npt.NDArray[np.floating]] = []
    for unit in rich.progress.track(
        units.itertuples(),
        description="getting spike counts for units",
        total=len(units),
    ):
        spikes_per_interval = get_spike_counts(
            np.array(unit.spike_times), np.array(_intervals)
        )
        if as_spikes_per_second:
            spikes_per_interval / np.diff(_intervals)
        _spikes_per_interval_per_unit.append(spikes_per_interval)

    spikes_per_interval_per_unit = apply_invalid_intervals(
        session, _intervals, _spikes_per_interval_per_unit, unit_selection
    )
    return spikes_per_interval_per_unit


def get_response_in_intervals(
    session,
    response_intervals: Interval | Iterable[Interval],
    baseline_intervals: Interval | Iterable[Interval],
    unit_selection: UnitSelection | None = None,
    as_spikes_per_second: bool = True,
    as_normalized_ratio: bool = False,
) -> npt.NDArray[np.floating]:
    if not as_spikes_per_second:
        assert np.diff(parse_intervals(response_intervals)) == np.diff(
            parse_intervals(baseline_intervals)
        ), "response and baseline intervals must have same duration to express response as spike counts"
    _response = get_spike_counts_in_intervals(
        session, response_intervals, unit_selection, as_spikes_per_second
    )
    _baseline = get_spike_counts_in_intervals(
        session, baseline_intervals, unit_selection, as_spikes_per_second
    )

    output = _response - _baseline
    if as_normalized_ratio:
        output /= _baseline
        # 0/0 is NaN, make it 0
        output[(_baseline == 0) & (_response == 0)] = 0
    return output


VIS_MIN_RESP_LATENCY = 0.025
AUD_MIN_RESP_LATENCY = 0.01
OPTO_MIN_RESP_LATENCY = 0.01

VIS_RESP_WINDOW = 0.2
AUD_RESP_WINDOW = 0.2
OPTO_RESP_WINDOW = 0.1


def add_opto_response_metric(session) -> None:
    """Adds `opto_response` metric to `session.units`"""
    trials = session.intervals.get(get_taskcontrol_intervals_table_name("OptoTagging"))
    if trials is None:
        return


def add_vis_response_metric(session) -> None:
    """Adds `vis_response` metric to `session.units`"""
    trials = session.intervals[
        get_taskcontrol_intervals_table_name("VisRFMapping")
    ].to_dataframe()
    if trials is None:
        return
    units = session.units[:]  # slight speed-up by creating df only once
    get_response_in_intervals(
        session,
        response_intervals=zip(
            (start := trials.stim_start_time[:] + VIS_MIN_RESP_LATENCY),
            start + VIS_RESP_WINDOW,
            strict=False,
        ),
        baseline_intervals=zip(
            trials.start_time[:], trials.stim_start_time[:], strict=False
        ),
        as_spikes_per_second=True,
        as_normalized_ratio=False,
        unit_selection=units,
    )
    trials = trials.query("is_small_field_grating")
    stim_by_unit_responses = []
    for grating_x in trials.grating_x.unique():
        for grating_y in trials.grating_y.unique():
            t = trials.query(f"grating_x == {grating_x} and grating_y == {grating_y}")
            if len(t) == 0:
                continue
            # get mean response for each unit to this stim location
            stim_by_unit_responses.append(
                np.nanmean(
                    get_response_in_intervals(
                        session,
                        response_intervals=zip(
                            (start := t.stim_start_time[:] + VIS_MIN_RESP_LATENCY),
                            start + VIS_RESP_WINDOW,
                            strict=False,
                        ),
                        baseline_intervals=zip(
                            t.start_time[:], t.stim_start_time[:], strict=False
                        ),
                        as_spikes_per_second=True,
                        as_normalized_ratio=False,
                        unit_selection=session.units[
                            :
                        ],  # slight speed-up by creating df only once
                    ),
                    axis=1,
                )
            )
    session.units.add_column(
        name="vis_response",
        description="mean change in firing rate in response to small-field grating stimulus at most responsive location on screen",
        data=np.nanmax(stim_by_unit_responses, axis=0),
    )


def add_aud_response_metric(session) -> None:
    """Adds `aud_response` metric to `session.units`"""
    trials = session.intervals["AudRFMapping"]
    if trials is None:
        return


if __name__ == "__main__":
    import doctest

    import dotenv

    dotenv.load_dotenv(dotenv.find_dotenv(usecwd=True))
    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )



from __future__ import annotations

import datetime
import importlib.metadata
import zoneinfo

import npc_io
import npc_lims
import npc_session


def get_package_version() -> str:
    return importlib.metadata.version("npc_sessions")


def is_stim_file(
    path: npc_io.PathLike,
    subject_spec: str | int | npc_session.SubjectRecord | None = None,
    date_spec: str | npc_session.DateRecord | None = None,
    time_spec: str | npc_session.TimeRecord | None = None,
) -> bool:
    """Does the string or path match a known stimulus file pattern.

    Optional arguments can be used to check the subject, date, and time.
    >>> is_stim_file("366122_20230101_120000.hdf5")
    True
    >>> is_stim_file("366122_20230101_120000.hdf5", date_spec="20230101")
    True
    >>> is_stim_file("366122.stim.pkl")
    True
    >>> is_stim_file("366123.stim.pkl", subject_spec="366122")
    False
    >>> is_stim_file("366122/test.stim.pkl", subject_spec="366122")
    True
    """
    path = npc_io.from_pathlike(path)
    subject_from_path = npc_session.extract_subject(path.as_posix())
    date_from_path = npc_session.extract_isoformat_date(path.as_posix())
    time_from_path = npc_session.extract_isoformat_time(path.as_posix())

    is_stim = False
    is_stim |= path.suffix in (".pkl",) and any(
        label in path.stem for label in ("stim", "mapping", "opto", "behavior")
    )
    is_stim |= (
        path.suffix in (".hdf5",) and bool(subject_from_path) and bool(date_from_path)
    )
    is_correct = True
    if subject_spec:
        is_correct &= subject_from_path == npc_session.SubjectRecord(str(subject_spec))
    if date_spec:
        is_correct &= date_from_path == npc_session.DateRecord(date_spec)
    if time_spec:
        is_correct &= time_from_path == npc_session.TimeRecord(time_spec)
    return is_stim and is_correct


def assert_s3_write_credentials() -> None:
    test = npc_lims.DR_DATA_REPO / "test.txt"
    test.touch()
    test.unlink()


def get_aware_dt(dt: str | datetime.datetime) -> datetime.datetime:
    """Add Seattle timezone info to a datetime string or object."""
    if isinstance(dt, datetime.datetime):
        dt = dt.isoformat()
    return npc_session.DatetimeRecord(dt).dt.replace(
        tzinfo=zoneinfo.ZoneInfo("America/Los_Angeles")
    )


def get_taskcontrol_intervals_table_name(taskcontrol_class_name: str) -> str:
    """Return the name of the taskcontrol interval.

    >>> get_taskcontrol_intervals_table_name("DynamicRouting1")
    'trials'
    >>> get_taskcontrol_intervals_table_name("SpontaneousRewards")
    'spontaneous_rewards_trials'
    >>> get_taskcontrol_intervals_table_name("OptoTagging")
    'optotagging_trials'
    >>> get_taskcontrol_intervals_table_name("VisRFMapping")
    'vis_rf_mapping_trials'
    """
    if taskcontrol_class_name == "DynamicRouting1":
        return "trials"
    # convert TitleCase to snake_case:
    snake = "".join(
        [
            c if c.islower() else f"_{c.lower()}"
            for c in taskcontrol_class_name
            if c.isalnum()
        ]
    ).lstrip("_")

    # acronyms/abbreviations are tricky to deal with - handling them on
    # case-by-case basis for now:
    snake = snake.replace("_r_f_mapping", "_rf_mapping")
    snake = snake.replace("opto_tagging", "optotagging")
    return snake + "_trials"


if __name__ == "__main__":
    import doctest

    import dotenv

    dotenv.load_dotenv(dotenv.find_dotenv(usecwd=True))
    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )



import tempfile

import pytest
import npc_sessions
import npc_sessions.aind_data_schema


@pytest.fixture(scope="module")
def ephys_session():
    return npc_sessions.DynamicRoutingSession('DRpilot_676909_20231213-08-03')


@pytest.fixture(scope="module")
def templeton_session():
    return npc_sessions.DynamicRoutingSession('628801_2022-09-20')


@pytest.fixture(scope="module")
def behavior_session():
    return npc_sessions.DynamicRoutingSession('715706_20240529')


@pytest.fixture(scope="module")
def all_sessions(ephys_session, templeton_session, behavior_session):
    return [behavior_session, ephys_session, templeton_session]

def test_data_description_metadata(all_sessions):
    for s in all_sessions:
        m = npc_sessions.aind_data_schema.get_data_description_model(s)
        m.write_standard_file(tempfile.mkdtemp())
    
def test_acquisition_metadata(all_sessions):
    for s in all_sessions:
        m = npc_sessions.aind_data_schema.get_acquisition_model(s)
        m.write_standard_file(tempfile.mkdtemp())
    
if __name__ == "__main__":
    import pytest
    pytest.main(['-s', __file__])


import contextlib
import datetime
import itertools
import json
import logging
import re
from collections.abc import Iterable
from typing import TypeGuard

import aind_data_schema.components.configs
import aind_data_schema.components.coordinates
import aind_data_schema.components.identifiers
import aind_data_schema.core.acquisition
import aind_data_schema_models.brain_atlas
import aind_data_schema_models.modalities
import aind_data_schema_models.stimulus_modality
import aind_data_schema_models.units
import numpy as np
import upath

from npc_sessions.aind_data_schema import instrument
from npc_sessions.sessions import DynamicRoutingSession, DynamicRoutingSurfaceRecording

logger = logging.getLogger(__name__)


def is_surface_recording(
    session: DynamicRoutingSession,
) -> TypeGuard[DynamicRoutingSurfaceRecording]:
    return isinstance(session, DynamicRoutingSurfaceRecording)


def _merge_data_streams(
    data_streams: Iterable[aind_data_schema.core.acquisition.DataStream],
) -> aind_data_schema.core.acquisition.DataStream:
    data_streams = tuple(data_streams)
    return aind_data_schema.core.acquisition.DataStream(
        stream_start_time=min(ds.stream_start_time for ds in data_streams),
        stream_end_time=max(ds.stream_end_time for ds in data_streams),
        modalities=sorted(
            {mod for ds in data_streams for mod in ds.modalities}, key=lambda m: m.name
        ),
        code=sorted(
            itertools.chain.from_iterable(ds.code for ds in data_streams if ds.code),
            key=lambda c: c.url,
        )
        or None,
        active_devices=sorted(
            itertools.chain.from_iterable(ds.active_devices for ds in data_streams),
        ),
        configurations=list(
            itertools.chain.from_iterable(ds.configurations for ds in data_streams)
        ),
    )


def get_acquisition_model(
    session: DynamicRoutingSession,
) -> aind_data_schema.core.acquisition.Acquisition:
    """Get the Pydantic model corresponding to the 'acquisition.json' for a given session."""
    subject_id = str(session.id.subject)
    acquisition_start_time = session.session_start_time
    if session.is_sync:
        acquisition_end_time = session.sync_data.stop_time
    elif session.epochs.stop_time:
        acquisition_end_time = session.session_start_time + datetime.timedelta(
            seconds=max(session.epochs.stop_time)
        )
    elif session.ephys_timing_data:  # eg surface channel recording
        acquisition_end_time = session.session_start_time + datetime.timedelta(
            seconds=session.ephys_timing_data[0].stop_time
        )
    else:
        raise ValueError(f"Unable to determine acquisition end time for {session}")
    if isinstance(session, DynamicRoutingSurfaceRecording):
        experimenters = session.main_recording.experimenter or ["unknown"]
    else:
        experimenters = session.experimenter or ["NSB trainer"]
    if is_surface_recording(session):
        instrument_id = session.main_recording.rig
        acquisition_type = "ecephys surface recording"
    else:
        instrument_id = session.rig
        keywords: list[str] = []
        if session.is_ephys:
            keywords.append("ecephys")
        if session.is_task:
            keywords.append("behavior")
        for k in [
            "naive",
            "injection_perturbation",
            "injection_control",
            "opto_perturbation",
            "opto_control",
        ]:
            if k in session.keywords:
                keywords.append(k)
        acquisition_type = " ".join(k.replace("_", " ") for k in keywords)
    coordinate_system = None
    data_streams = _get_data_streams(session)
    stimulus_epochs = _get_stimulus_epochs(session)
    if is_surface_recording(session):
        subject_details = None
    else:
        subject_details = aind_data_schema.core.acquisition.AcquisitionSubjectDetails(
            mouse_platform_name="Brain Observatory running disc",
            reward_consumed_total=(
                (np.nanmean(session.sam.rewardSize) * len(session.sam.rewardTimes))
                if session.is_task
                else None
            ),
            reward_consumed_unit=aind_data_schema_models.units.VolumeUnit.ML,
        )
    return aind_data_schema.core.acquisition.Acquisition(
        subject_id=subject_id,
        acquisition_start_time=acquisition_start_time,
        acquisition_end_time=acquisition_end_time,
        experimenters=experimenters,
        protocol_id=None,
        ethics_review_id=["2104"],
        instrument_id=instrument_id,
        acquisition_type=acquisition_type,
        notes=session.notes,
        coordinate_system=coordinate_system,
        data_streams=[_merge_data_streams(data_streams)],
        stimulus_epochs=stimulus_epochs,
        subject_details=subject_details,
    )


def get_active_devices(script_name: str, session: DynamicRoutingSession) -> list[str]:
    stim = aind_data_schema_models.stimulus_modality.StimulusModality
    modalities = get_modalities(script_name, session)
    device_names = [instrument.TASKCONTROL_DAQ.name, instrument.CAMSTIM_DAQ.name]
    if stim.VISUAL in modalities:
        device_names.append(instrument.MONITOR.name)
        if session.is_sync:
            device_names.extend([instrument.PHOTODIODE.name])
    if stim.AUDITORY in modalities:
        device_names.append(instrument.SPEAKER.name)
        if session.is_sync:
            device_names.extend([instrument.MICROPHONE.name])
    if stim.OPTOGENETICS in modalities:
        device_names.extend(
            [
                instrument.LASER_488.name,
                instrument.LASER_633.name,
                instrument.LASER_GALVO_X.name,
                instrument.LASER_GALVO_Y.name,
                instrument.OPTO_DAQ.name,
            ]
        )
    if session.is_sync:
        device_names.append(instrument.SYNC_DAQ.name)

    return sorted(set(device_names))


def get_modalities(
    script_name: str,
    session: DynamicRoutingSession,
) -> list[aind_data_schema_models.stimulus_modality.StimulusModality]:
    stim = aind_data_schema_models.stimulus_modality.StimulusModality
    modalities = []
    if any(
        name in script_name
        for name in (
            "DynamicRouting",
            "RFMapping",
            "LuminanceTest",
            "Spontaneous",
        )
    ):
        modalities.append(stim.VISUAL)
    if any(name in script_name for name in ("DynamicRouting", "RFMapping")):
        modalities.append(stim.AUDITORY)
    if session.is_opto and any(name in script_name for name in ("DynamicRouting",)):
        modalities.append(stim.OPTOGENETICS)
    if any(name in script_name for name in ("OptoTagging",)):
        modalities.append(stim.OPTOGENETICS)
    return modalities or [stim.NO_STIMULUS]


def _get_stimulus_epochs(
    session: DynamicRoutingSession,
) -> list[aind_data_schema.core.acquisition.StimulusEpoch]:
    if is_surface_recording(session):
        return []

    def get_speaker_config(
        script_name: str,
    ) -> aind_data_schema.components.configs.SpeakerConfig | None:
        stim = aind_data_schema_models.stimulus_modality.StimulusModality
        if stim.AUDITORY not in get_modalities(script_name, session):
            return None
        return aind_data_schema.components.configs.SpeakerConfig(
            device_name="Speaker",
            volume=68.0,
            volume_unit="decibels",
        )

    def get_performance_metrics(
        script_name: str,
    ) -> aind_data_schema.core.acquisition.PerformanceMetrics | None:
        if "DynamicRouting" not in script_name or not session.is_task:
            return None
        block_metrics = {}
        for block_index, (
            hit_count,
            dprime_same_modal,
            dprime_other_modal_go,
            block_stim_rewarded,
        ) in enumerate(
            zip(
                [int(v) for v in session.sam.hitCount],
                [float(v) for v in session.sam.dprimeSameModal],
                [float(v) for v in session.sam.dprimeOtherModalGo],
                [str(v) for v in session.sam.blockStimRewarded],
                strict=False,
            )
        ):
            block_metrics[str(block_index)] = dict(
                block_index=block_index,
                block_stim_rewarded=block_stim_rewarded,
                hit_count=hit_count,
                dprime_same_modal=dprime_same_modal,
                dprime_other_modal_go=dprime_other_modal_go,
            )
        return aind_data_schema.core.acquisition.PerformanceMetrics(
            output_parameters={
                "block_metrics": block_metrics,
                "task_version": session.sam.taskVersion,
            },
            reward_consumed_during_epoch=np.nanmean(session.sam.rewardSize)
            * sum(session.trials[:].is_rewarded),
            reward_consumed_unit=aind_data_schema_models.units.VolumeUnit.ML,
            trials_total=len(session.trials[:]),
            trials_rewarded=sum(session.trials[:].is_contingent_reward),
        )

    def get_laser_configs(
        script_name: str,
    ) -> list[aind_data_schema.components.configs.LaserConfig] | None:
        configs = []
        for laser in [instrument.LASER_488, instrument.LASER_633]:
            if laser.name not in get_active_devices(script_name, session):
                continue
            if script_name == "OptoTagging":
                column_name = "power"
            elif script_name == "DynamicRouting1":
                column_name = "opto_power"
            else:
                raise NotImplementedError(
                    f"Unknown script name {script_name}: unsure how to get laser power from intervals table"
                )
            trials = next(v for k, v in session._all_trials.items() if script_name in k)
            max_power = np.nanmax(getattr(trials, column_name))
            if np.isnan(max_power):  # control session
                max_power = 0.0
            configs.append(
                aind_data_schema.components.configs.LaserConfig(
                    device_name=laser.name,
                    wavelength=laser.wavelength,
                    wavelength_unit=laser.wavelength_unit,
                    power=max_power,
                    power_unit=aind_data_schema_models.units.PowerUnit.MW,
                    power_measured_at="Brain surface",
                )
            )
        return configs or None

    def get_version() -> str | None:
        if "blob/main" in session.source_script:
            return None
        return (
            session.source_script.split("DynamicRoutingTask/")[-1]
            .split("/DynamicRouting1.py")[0]
            .strip("/")
        )

    def get_url(epoch_name: str) -> str:
        return session.source_script.replace(
            "DynamicRouting1", get_taskcontrol_file(epoch_name)
        )

    def get_taskcontrol_file(epoch_name: str) -> str:
        if upath.UPath(
            session.source_script.replace("DynamicRouting1", epoch_name)
        ).exists():
            return epoch_name
        return "TaskControl"

    def get_code(script_name: str) -> aind_data_schema.core.acquisition.Code:
        return aind_data_schema.core.acquisition.Code(
            url=get_url(script_name),
            version=get_version(),
            container=None,
            language="Python",
            language_version="3.9",
            core_dependency=aind_data_schema.components.identifiers.Software(
                name="PsychoPy",
                version="2022.1.2",
            ),
        )

    def get_configurations(
        script_name: str,
    ) -> list[aind_data_schema.components.configs.DeviceConfig]:
        configurations = []
        speaker_config = get_speaker_config(script_name)
        if speaker_config:
            configurations.append(speaker_config)
        laser_configs = get_laser_configs(script_name)
        if laser_configs:
            configurations.extend(laser_configs)
        return configurations

    def get_training_protocol_name(
        script_name: str,
        session: DynamicRoutingSession,
    ) -> str | None:
        if "DynamicRouting" not in script_name or not session.is_task:
            return None
        protocol = "dynamic_routing"
        if session.is_context_naive:
            protocol += "_context_naive"
        if session.is_naive:
            protocol += "_naive"
        return protocol

    def get_curriculum_status(
        script_name: str,
        session: DynamicRoutingSession,
    ) -> str | None:
        if "DynamicRouting" not in script_name or not session.is_task:
            return None
        # extract 'stage X'
        stages = re.findall(
            r"stage\s?(\d+)",
            session.sam.taskVersion.lower(),
        )
        if not stages:
            return None
        return f"stage {stages[0]}"

    aind_epochs = []
    for nwb_epoch in session.epochs:
        script_name = nwb_epoch.script_name.item()
        aind_epochs.append(
            aind_data_schema.core.acquisition.StimulusEpoch(
                stimulus_start_time=datetime.timedelta(
                    seconds=nwb_epoch.start_time.item()
                )
                + session.session_start_time,
                stimulus_end_time=datetime.timedelta(seconds=nwb_epoch.stop_time.item())
                + session.session_start_time,
                stimulus_name=script_name,
                code=get_code(script_name),
                stimulus_modalities=get_modalities(script_name, session),
                performance_metrics=get_performance_metrics(script_name),
                notes=nwb_epoch.notes.item(),
                active_devices=get_active_devices(script_name, session),
                configurations=get_configurations(script_name),
                training_protocol_name=get_training_protocol_name(script_name, session),
                curriculum_status=get_curriculum_status(script_name, session),
            )
        )
    return aind_epochs


def _get_data_streams(
    session: DynamicRoutingSession,
) -> list[aind_data_schema.core.acquisition.DataStream]:
    def get_core_dependency(
        stream_name: str,
    ) -> aind_data_schema.core.acquisition.Code | None:
        match stream_name:
            case "Ephys":
                assert (
                    session.is_ephys
                ), "Ephys dependency info requested for non-ephys session - should not be possible"
                try:
                    version = session.ephys_settings_xml_data.open_ephys_version
                except (AttributeError, FileNotFoundError):
                    version = "0.6.6"  # most-commonly used version
                return aind_data_schema.components.identifiers.Software(
                    name="Open Ephys GUI",
                    version=version,
                )
            case "Camstim":
                return aind_data_schema.components.identifiers.Software(
                    name="PsychoPy",
                    version="2022.1.2",
                )

            case "MVR":
                assert (
                    session.is_video
                ), "MVR dependency info requested for non-video session - should not be possible"
                try:
                    version = next(iter(session.mvr.info_data.values()))["MVR Version"]
                except (StopIteration, AttributeError, KeyError, FileNotFoundError):
                    version = "1.1.7"  # most-commonly used version
                return aind_data_schema.components.identifiers.Software(
                    name="MultiVideoRecorder",
                    version=version,
                )
            case "Sync":
                return aind_data_schema.components.identifiers.Software(
                    name="nidaqmx",
                    version="0.6.2",
                )
            case _:
                return None

    def get_np_services_code(
        stream_name: str,
    ) -> list[aind_data_schema.core.acquisition.Code] | None:
        if stream_name not in ("MVR", "Sync", "Camstim", "Ephys"):
            raise ValueError(f"Unknown stream name {stream_name} for NP services code")
        script_name = {
            "MVR": "proxies",
            "Sync": "proxies",
            "Camstim": "proxies",
            "Ephys": "open_ephys",
        }
        url = f"https://github.com/AllenInstitute/np_services/blob/main/src/np_services/{script_name[stream_name]}.py"
        return [
            aind_data_schema.core.acquisition.Code(
                url=url,
                container=None,
                language="Python",
                language_version="3.9",
                core_dependency=get_core_dependency(stream_name),
            )
        ]

    data_streams = []
    modality = aind_data_schema_models.modalities.Modality

    if session.is_sync:
        data_streams.append(
            aind_data_schema.core.acquisition.DataStream(
                stream_start_time=session.sync_data.start_time,
                stream_end_time=session.sync_data.stop_time,
                modalities=[modality.BEHAVIOR, modality.BEHAVIOR_VIDEOS],
                code=get_np_services_code("Sync"),
                active_devices=[instrument.SYNC_DAQ.name],
                configurations=[],
            )
        )
    if session.stim_paths and len(session.epochs.script_name):
        data_streams.append(
            aind_data_schema.core.acquisition.DataStream(
                stream_start_time=session.session_start_time
                + datetime.timedelta(seconds=min(session.epochs.stop_time)),
                stream_end_time=session.session_start_time
                + datetime.timedelta(seconds=max(session.epochs.stop_time)),
                modalities=[modality.BEHAVIOR],
                active_devices=sorted(
                    set(
                        itertools.chain.from_iterable(
                            get_active_devices(s, session)
                            for s in session.epochs.script_name
                        )
                    )
                ),
                code=get_np_services_code("Camstim"),
                configurations=[],
            )
        )
    if session.is_video:
        if session.is_sync:
            active_cameras = [
                instrument.EYE_CAMERA,
                instrument.FRONT_CAMERA,
                instrument.SIDE_CAMERA,
                instrument.NOSE_CAMERA,
            ]
        else:
            active_cameras = [instrument.SIDE_CAMERA]
        data_streams.append(
            aind_data_schema.core.acquisition.DataStream(
                stream_start_time=session.session_start_time
                + datetime.timedelta(
                    seconds=min(
                        np.nanmin(times.timestamps)
                        for times in session._video_frame_times
                    )  # min frame time across all vids
                ),
                stream_end_time=session.session_start_time
                + datetime.timedelta(
                    seconds=max(
                        np.nanmax(times.timestamps)
                        for times in session._video_frame_times
                    )  # max frame time across all vids
                ),
                modalities=[modality.BEHAVIOR_VIDEOS],
                active_devices=[cam.name for cam in active_cameras],
                code=get_np_services_code("MVR"),
                configurations=[],
            )
        )
    if session.is_ephys:
        # Build manipulator configurations for each probe

        if is_surface_recording(session):
            rec = session.main_recording
        else:
            rec = session

        if (
            path := next(
                (p for p in rec.raw_data_paths if p.name.endswith("_dye.json")), None
            )
        ) is not None:
            dye = json.loads(path.read_text())["dye"]
        else:
            dye = "unknown"

        ephys_configs = []
        for probe in session.probe_letters_to_use:
            translation = []
            # a few sessions didn't have newscale logging enabled: no way to get their coords
            with contextlib.suppress(AttributeError):
                row = rec._manipulator_positions[:].query(
                    f"electrode_group_name == '{probe.name}'"
                )
                if not row.empty:
                    translation = [row[axis].item() for axis in ["x", "y", "z"]]

            manipulator_config = aind_data_schema.components.configs.ManipulatorConfig(
                device_name=probe.id,
                coordinate_system=aind_data_schema.components.coordinates.CoordinateSystem(
                    name=f"{probe.name} XYZ",
                    origin=aind_data_schema.components.coordinates.Origin.TIP,
                    axes=[
                        aind_data_schema.components.coordinates.Axis(
                            name=aind_data_schema.components.coordinates.AxisName.X,
                            direction=aind_data_schema.components.coordinates.Direction.LR,
                        ),
                        aind_data_schema.components.coordinates.Axis(
                            name=aind_data_schema.components.coordinates.AxisName.Y,
                            direction=aind_data_schema.components.coordinates.Direction.BF,
                        ),
                        aind_data_schema.components.coordinates.Axis(
                            name=aind_data_schema.components.coordinates.AxisName.Z,
                            direction=aind_data_schema.components.coordinates.Direction.SI,
                        ),
                    ],
                    axis_unit=aind_data_schema.components.coordinates.SizeUnit.UM,
                ),
                local_axis_positions=aind_data_schema.components.coordinates.Translation(
                    translation=translation,
                ),
            )
            match probe:
                case "A":
                    primary_targeted_structure = (
                        aind_data_schema_models.brain_atlas.CCFv3.ROOT
                    )
                case "B":
                    primary_targeted_structure = (
                        aind_data_schema_models.brain_atlas.CCFv3.ROOT
                    )
                case "C":
                    primary_targeted_structure = (
                        aind_data_schema_models.brain_atlas.CCFv3.ROOT
                    )
                case "D":
                    primary_targeted_structure = (
                        aind_data_schema_models.brain_atlas.CCFv3.ROOT
                    )
                case "E":
                    primary_targeted_structure = (
                        aind_data_schema_models.brain_atlas.CCFv3.ROOT
                    )
                case "F":
                    primary_targeted_structure = (
                        aind_data_schema_models.brain_atlas.CCFv3.ROOT
                    )

            probe_config = aind_data_schema.components.configs.ProbeConfig(
                device_name=probe.name,
                primary_targeted_structure=primary_targeted_structure,
                coordinate_system=manipulator_config.coordinate_system,
                transform=[manipulator_config.local_axis_positions],
                dye=dye,
            )
            ephys_assembly = aind_data_schema.components.configs.EphysAssemblyConfig(
                device_name=probe.id,
                manipulator=manipulator_config,
                probes=[probe_config],
                modules=None,
            )
            ephys_configs.append(ephys_assembly)
        data_streams.append(
            aind_data_schema.core.acquisition.DataStream(
                stream_start_time=session.session_start_time
                + datetime.timedelta(
                    seconds=min(
                        timing.start_time for timing in session.ephys_timing_data
                    )
                ),
                stream_end_time=session.session_start_time
                + datetime.timedelta(
                    seconds=max(
                        timing.stop_time for timing in session.ephys_timing_data
                    )
                ),
                modalities=[modality.ECEPHYS],
                active_devices=[
                    probe.name for probe in session.probe_letters_to_use
                ],  # keep names synced with 'instrument.py'
                configurations=ephys_configs,
                code=get_np_services_code("Ephys"),
            )
        )
    return data_streams


if __name__ == "__main__":
    session = DynamicRoutingSession("814666_20251107")
    metadata = get_acquisition_model(session)
    print(metadata.model_dump_json(indent=2))
    with open("acquisition_814666_2025-11-07.json", "w") as f:
        f.write(metadata.model_dump_json(indent=2))
    # session = DynamicRoutingSession('628801_2022-09-20')
    # metadata = get_acquisition_model(session)
    # print(metadata.model_dump_json(indent=2))
    # with open('acquisition_628801_2022-09-20.json', 'w') as f:
    #     f.write(metadata.model_dump_json(indent=2))



import logging

import aind_data_schema.components.identifiers
import aind_data_schema.core.data_description
import aind_data_schema_models.data_name_patterns
import aind_data_schema_models.modalities
import aind_data_schema_models.organizations

from npc_sessions.sessions import DynamicRoutingSession

logger = logging.getLogger(__name__)


def get_data_description_model(
    session: DynamicRoutingSession,
) -> aind_data_schema.core.data_description.DataDescription:
    """Get the Pydantic model corresponding to the 'data_description.json' for a given session."""
    subject_id = str(session.id.subject)
    creation_time = session.session_start_time
    platform = "ecephys" if session.is_ephys else "behavior"
    name = f"{platform}_{subject_id}_{creation_time.strftime('%Y-%m-%d_%H-%M-%S')}"

    investigators = [
        aind_data_schema.components.identifiers.Person(
            name="Shawn Olsen",
            registry=aind_data_schema.components.identifiers.Registry.ORCID,
            registry_identifier="0000-0002-9568-7057",
        ),
        # others added below depending on project
    ]

    if session.is_templeton:
        project_name = "Templeton"
        funding_source = aind_data_schema.core.data_description.Funding(
            funder=aind_data_schema_models.organizations.Organization.TWCF,
        )
        investigators.append(
            aind_data_schema.components.identifiers.Person(
                name="Ethan McBride",
                registry=aind_data_schema.components.identifiers.Registry.ORCID,
                registry_identifier="0000-0001-9489-2828",
            )
        )
    else:
        project_name = "Dynamic Routing"
        funding_source = aind_data_schema.core.data_description.Funding(
            funder=aind_data_schema_models.organizations.Organization.AI,
        )
        investigators.append(
            aind_data_schema.components.identifiers.Person(
                name="Corbett Bennett",
                registry=aind_data_schema.components.identifiers.Registry.ORCID,
                registry_identifier="0009-0001-2847-7754",
            )
        )

    modalities = []
    if session.is_ephys:
        modalities.append(aind_data_schema_models.modalities.Modality.ECEPHYS)
    if session.is_task:
        modalities.append(aind_data_schema_models.modalities.Modality.BEHAVIOR)
    if session.is_video:
        modalities.append(aind_data_schema_models.modalities.Modality.BEHAVIOR_VIDEOS)

    tags: list[str] = []
    if session.is_surface_channels:
        tags.append("surface_recording")
    if session.is_hab:
        tags.append("hab")
    if session.is_opto:
        tags.append("opto_perturbation")
    elif session.is_opto_control:
        tags.append("opto_control")
    if session.is_production:
        tags.append("prod")
    else:
        tags.append("dev")
    if session.is_injection_perturbation:
        tags.append("injection_perturbation")
    elif session.is_injection_control:
        tags.append("injection_control")
    if session.is_context_naive:
        tags.append("context_naive")
    if session.is_stage_5_passed:
        tags.append("stage_5_passed")
    if session.is_task and session.is_first_block_aud:
        tags.append("first_block_aud")
    elif session.is_task and session.is_first_block_vis:
        tags.append("first_block_vis")
    if session.is_task and session.is_late_autorewards:
        tags.append("late_autorewards")
    elif session.is_task:
        tags.append("early_autorewards")

    return aind_data_schema.core.data_description.DataDescription(
        name=name,
        institution=aind_data_schema_models.organizations.Organization.AIND,
        funding_source=[funding_source],
        data_level=aind_data_schema_models.data_name_patterns.DataLevel.RAW,
        group=aind_data_schema_models.data_name_patterns.Group.EPHYS,
        subject_id=subject_id,
        creation_time=creation_time,
        investigators=investigators,
        project_name=project_name,
        restrictions=None,
        modalities=modalities,
        tags=tags,
        data_summary=session.experiment_description,
    )


if __name__ == "__main__":
    # Example usage
    session = DynamicRoutingSession("814666_20251107")
    data_description = get_data_description_model(session)
    print(data_description.model_dump_json(indent=2))
    session = DynamicRoutingSession("628801_2022-09-20")
    data_description = get_data_description_model(session)
    print(data_description.model_dump_json(indent=2))



from __future__ import annotations

import collections.abc
import datetime
import doctest
import functools
import logging
import reprlib
from collections.abc import Generator, Iterator, Sequence
from typing import Any, Literal

import npc_io
import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)

EXCLUDE_IF_ONLY_NANS = True
"""If True, exclude columns with all NaN values from the resulting dict. This is
the behavior expected from nwb validation. For some columns that are not always
available (such as `opto_location_name`) we return `np.full(np.nan)`: if those
columns aren't dropped, then dataframe schemas will become inconsistent (str vs double)"""

EXPLODE_LISTS = True
"""If True, any with values in the resulting dataframe that are list-like (list, tuple, set, np.ndarray), will be
expanded into separate rows, e.g. 

this dataframe:
   start_time  stop_time  trial_idx opto_bregma_x opto_bregma_y                opto_target
0        12.3       13.3          3     [-10, 10]       [2, -2]  [upper left, lower right]
1        15.7       16.7          4       [-5, 5]       [-4, 4]  [lower left, upper right]

becomes:
   start_time  stop_time  trial_idx opto_bregma_x opto_bregma_y  opto_target
0        12.3       13.3          3           -10             2   upper left
1        12.3       13.3          3            10            -2  lower right
2        15.7       16.7          4            -5            -4   lower left
3        15.7       16.7          4             5             4  upper right

via `df.explode(['opto_bregma_x', 'opto_bregma_y', 'opto_target'], ignore_index=True)`

The major assumption here is that all list-like values in the dataframe are of the
same length, and correspond to the same parameter.

This can make it easier to deal with trials with multiple values for a single
location parameter, such as bilateral opto stimulation targets
"""


class PropertyDict(collections.abc.Mapping):
    """Dict type, where the keys are the class's properties (regular attributes
    and property getters) which don't have a leading underscore, and values are
    corresponding property values.

    Methods can be added and used as normal, they're just not visible in the
    dict.
    """

    @property
    def _properties(self) -> tuple[str, ...]:
        """Names of properties without leading underscores. No methods.
        Excludes properties with all NaN values."""
        self_and_super_attrs = tuple(
            attr
            for cls in reversed(self.__class__.__mro__)
            for attr in cls.__dict__.keys()
        )
        dict_attrs = collections.abc.Mapping.__dict__.keys()
        no_dict_attrs = tuple(
            attr for attr in self_and_super_attrs if attr not in dict_attrs
        )
        no_leading_underscore = tuple(attr for attr in no_dict_attrs if attr[0] != "_")
        no_functions = tuple(
            attr for attr in no_leading_underscore if not callable(getattr(self, attr))
        )

        def all_nan(attr):
            return (
                isinstance(a := getattr(self, attr), np.ndarray)
                and a.dtype == np.floating
                and all(np.isnan(a))
            )

        if EXCLUDE_IF_ONLY_NANS:
            return tuple(attr for attr in no_functions if not all_nan(attr))
        return no_functions

    @property
    def _dict(self) -> dict[str, Any]:
        return {attr: getattr(self, attr) for attr in self._properties}

    def __getitem__(self, key) -> Any:
        return self._dict[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._dict)

    def __len__(self) -> int:
        return len(self._dict)

    def __repr__(self) -> str:
        return reprlib.repr(self._dict)

    def to_add_trial_column(
        self,
    ) -> Generator[dict[Literal["name", "description"], str], None, None]:
        """Name and description for each trial column.

        Does not include `start_time` and `stop_time`: those columns
        always exist in an nwb trials table, trying to add them again raises
        an error.

        Iterate over result and unpack each dict:

        >>> for column in obj.to_add_trial_column(): # doctest: +SKIP
        ...    nwb_file.add_trial_column(**column)

        """
        attrs = self._properties
        descriptions = self._docstrings
        missing = tuple(column for column in attrs if column not in descriptions)
        if any(missing):
            logger.warning(
                f"These properties do not have descriptions (add docstrings to their property getters): {missing}"
            )
        descriptions.update(dict(zip(missing, ("" for _ in missing), strict=False)))
        descriptions.pop("start_time", None)
        descriptions.pop("stop_time", None)
        return (
            {"name": name, "description": description}
            for name, description in descriptions.items()
        )

    def to_add_trial(
        self,
    ) -> Generator[dict[str, int | float | str | datetime.datetime], None, None]:
        """Column name and value for each trial.

        Iterate over result and unpack each dict:

        >>> for trial in obj.to_add_trial(): # doctest: +SKIP
        ...    nwb_file.add_trial(**trial)

        """
        mandatory_columns = ("start_time", "stop_time")
        if any(mandatory not in self._properties for mandatory in mandatory_columns):
            raise AttributeError(
                f"{self} is missing one of {mandatory_columns = } required for nwb trials table"
            )

        for trial in self.to_dataframe().iterrows():
            yield dict(trial[1])

    def to_dataframe(self) -> pd.DataFrame:
        """pandas dataframe"""
        df = pd.DataFrame(
            data={
                k: pd.Series(list(v) if isinstance(v, np.ndarray) else v)
                for k, v in self.items()
            }
        )
        if EXPLODE_LISTS:
            explode_cols = list(
                col
                for col in df.columns
                if isinstance(df[col][0], (list, tuple, set, np.ndarray))
            )
            if explode_cols:
                # all explode_cols should have values with the same dimensions -
                # pandas will raise an error if they don't
                df = df.explode(explode_cols, ignore_index=True)
        for col in df.columns:
            if df[col].dtype == object:
                if any(isinstance(v, str) for v in df[col].values):
                    df[col] = df[col].fillna("")
                elif all(isinstance(v, int) for v in df[col].values if not np.isnan(v)):
                    # cast to float
                    df[col] = df[col].apply(
                        lambda x: np.float64(x) if isinstance(x, int) else np.nan
                    )
        return df

    @property
    def _docstrings(self) -> dict[str, str]:
        """Docstrings of property getter methods that have no leading
        underscore in their name.

        - getting the docstring of a regular property/attribute isn't easy:
        if we query its docstring in the same way as a property getter/method,
        we'll just receive the docstring for its value's type.
        """

        def cls_attr(attr):
            return getattr(self.__class__, attr)

        {
            attr: ""
            for attr in self._properties
            if not isinstance(
                cls_attr(attr),
                (property, functools.cached_property, npc_io.cached_property),
            )
        }
        property_getters = {
            attr: cls_attr(attr).__doc__ or ""
            for attr in self._properties
            if isinstance(
                cls_attr(attr),
                (property, functools.cached_property, npc_io.cached_property),
            )
        }

        def fmt(docstring):
            return (
                docstring.replace("\n", " ")
                .replace("  ", "")
                .replace(".- ", "; ")
                .replace("- ", "; ")
                .replace(" ; ", "; ")
                .strip(". ")
            )

        return {
            attr: (
                fmt(cls_attr(attr).__doc__) if cls_attr(attr).__doc__ else ""
            )  # if no docstring present, __doc__ is None
            for attr in property_getters
        }


class TestPropertyDict(PropertyDict):
    """
    >>> obj = TestPropertyDict()
    >>> obj
    {'no_docstring': True, 'visible_property': True, 'visible_property_getter': True}

    >>> obj.invisible_method()
    True

    >>> obj._docstrings
    {'visible_property_getter': 'Docstring available', 'no_docstring': ''}
    """

    visible_property = True

    @property
    def visible_property_getter(self):
        """Docstring available"""
        return True

    @property
    def no_docstring(self):
        return True

    _invisible_property: Any = None

    def invisible_method(self):
        """Docstring not available"""
        return True


class TestPropertyDictInheritance(TestPropertyDict):
    """
    >>> obj = TestPropertyDictInheritance()
    >>> obj
    {'no_docstring': True, 'visible_property': True, 'visible_property_getter': True}
    """

    pass


class TestPropertyDictExports(PropertyDict):
    """
    >>> obj = TestPropertyDictExports()

    >>> for kwargs in obj.to_add_trial_column():
    ...     print(kwargs)
    {'name': 'test_start_time', 'description': 'Start of trial'}
    {'name': 'test_stop_time', 'description': 'End of trial'}

    note: `start_time` and `stop_time` cannot be added as columns (they always
    exist)

    >>> for kwargs in obj.to_add_trial():
    ...     print(kwargs)
    {'start_time': 0.0, 'stop_time': 1.0, 'test_start_time': 1.0, 'test_stop_time': 1.5}
    {'start_time': 1.0, 'stop_time': 2.0, 'test_start_time': 2.0, 'test_stop_time': 2.5}
    """

    start_time = [0, 1]
    stop_time = [1, 2]

    @property
    def test_start_time(self) -> Sequence[float]:
        "Start of trial"
        return [1.0, 2.0]

    @property
    def test_stop_time(self) -> Sequence[float]:
        "End of trial"
        return [start + 0.5 for start in self.test_start_time]


if __name__ == "__main__":
    doctest.testmod()



import contextlib
import logging
import re
from typing import Literal

import aind_data_schema.components.connections
import aind_data_schema.components.coordinates
import aind_data_schema.components.devices
import aind_data_schema.components.identifiers
import aind_data_schema.core.instrument
import aind_data_schema_models.coordinates
import aind_data_schema_models.devices
import aind_data_schema_models.modalities
import aind_data_schema_models.organizations
import aind_data_schema_models.units
from aind_data_schema.components.coordinates import Axis, AxisName, Direction

from npc_sessions.sessions import DynamicRoutingSession

logger = logging.getLogger(__name__)


RIG_COORDINATE_SYSTEM = aind_data_schema.components.coordinates.CoordinateSystem(
    name="BREGMA_PRU",
    origin=aind_data_schema.components.coordinates.Origin.BREGMA,
    axes=[
        Axis(name=AxisName.X, direction=Direction.PA),
        Axis(name=AxisName.Y, direction=Direction.RL),
        Axis(name=AxisName.Z, direction=Direction.DU),
    ],
    axis_unit=aind_data_schema.components.coordinates.SizeUnit.M,
)
MONITOR_COORDINATE_SYSTEM = aind_data_schema.components.coordinates.CoordinateSystem(
    name="MINDSCOPE_COPA_MONITOR_RDB",
    origin=aind_data_schema.components.coordinates.Origin.FRONT_CENTER,
    axes=[
        Axis(name=AxisName.X, direction=Direction.RL),
        Axis(name=AxisName.Y, direction=Direction.DU),
        Axis(name=AxisName.Z, direction=Direction.BF),
    ],
    axis_unit=aind_data_schema.components.coordinates.SizeUnit.M,
)
CAMERA_COORDINATE_SYSTEM = aind_data_schema.components.coordinates.CoordinateSystem(
    name="MINDSCOPE_COPA_CAMERA_LUB",
    origin=aind_data_schema.components.coordinates.Origin.FRONT_CENTER,
    axes=[
        Axis(name=AxisName.X, direction=Direction.LR),
        Axis(name=AxisName.Y, direction=Direction.UD),
        Axis(name=AxisName.Z, direction=Direction.BF),
    ],
    axis_unit=aind_data_schema.components.coordinates.SizeUnit.M,
)


def _normalize_rig_name(rig_name: str) -> str:
    return rig_name.replace(".", "").replace("-", "").lower()


def is_np_rig(rig_name: str) -> bool:
    return _normalize_rig_name(rig_name).startswith("np")


def is_behavior_box(rig_name: str) -> bool:
    return _normalize_rig_name(rig_name).startswith("b")


def is_og_rig(rig_name: str) -> bool:
    return _normalize_rig_name(rig_name).startswith("og")


def get_location(rig_name: str) -> str:
    """Get a Location object for a given rig name."""
    rig_name = _normalize_rig_name(rig_name)
    if rig_name.startswith("np"):
        return {
            "np0": "325",
            "np1": "325",
            "np2": "327",
            "np3": "342",
        }[rig_name]
    if rig_name == "og1":
        return "342"
    if rig_name.startswith("beh"):
        return "NSB"
    else:
        return "342"  # behavior B


DISC = aind_data_schema.components.devices.Disc(
    name="Brain Observatory running disc",
    radius="4.69",
    radius_unit="centimeter",
    output=aind_data_schema_models.devices.DaqChannelType.DO,
    surface_material="rubber",
    notes="Radius is distance from center to subject",
)

MONITOR = aind_data_schema.components.devices.Monitor(
    name="Stimulus monitor",
    manufacturer=aind_data_schema_models.organizations.Organization.ASUS,
    model="PA248",
    refresh_rate=60,
    width=1920,
    height=1200,
    size_unit=aind_data_schema_models.units.SizeUnit.PX,
    viewing_distance=15.3,
    viewing_distance_unit=aind_data_schema_models.units.SizeUnit.CM,
    brightness=43,
    contrast=50,
    coordinate_system=MONITOR_COORDINATE_SYSTEM,
    relative_position=[
        aind_data_schema_models.coordinates.AnatomicalRelative.RIGHT,
        aind_data_schema_models.coordinates.AnatomicalRelative.ANTERIOR,
        aind_data_schema_models.coordinates.AnatomicalRelative.SUPERIOR,
    ],
    transform=[
        aind_data_schema.components.coordinates.Affine(
            affine_transform=[
                [-0.80914, -0.58761, 0],
                [-0.12391, 0.17063, 0.97751],
                [-0.5744, 0.79095, -0.21087],
                [0.08751, -0.12079, 0.02298],
            ]
        ),
    ],
)


SPEAKER = aind_data_schema.components.devices.Speaker(
    name="Stimulus speaker",
    manufacturer=aind_data_schema_models.organizations.Organization.ISL,
    model="SPK-I-81345",
    coordinate_system=MONITOR_COORDINATE_SYSTEM,  # same system as monitor
    relative_position=[
        aind_data_schema_models.coordinates.AnatomicalRelative.RIGHT,
        aind_data_schema_models.coordinates.AnatomicalRelative.ANTERIOR,
        aind_data_schema_models.coordinates.AnatomicalRelative.SUPERIOR,
    ],
    transform=[
        aind_data_schema.components.coordinates.Affine(
            affine_transform=[
                [-0.82783, -0.4837, -0.28412],
                [-0.55894, 0.75426, 0.34449],
                [0.04767, 0.44399, -0.89476],
                [-0.00838, -0.09787, 0.18228],
            ]
        ),
    ],
)

LICK_SPOUT_ASSEMBLY = aind_data_schema.components.devices.LickSpoutAssembly(
    name="Reward delivery assembly",
    lick_spouts=[
        aind_data_schema.components.devices.LickSpout(
            name="Reward spout",
            manufacturer=aind_data_schema_models.organizations.Organization.HAMILTON,
            model="8649-01",
            spout_diameter=0.672,
            spout_diameter_unit=aind_data_schema_models.units.SizeUnit.MM,
            notes="Spout diameter is inner. Outer diameter is 1.575 mm.",
            solenoid_valve=aind_data_schema.components.devices.Device(
                name="Solenoid valve",
                manufacturer=aind_data_schema_models.organizations.Organization.NRESEARCH_INC,
                model="161K011",
            ),
            lick_sensor=aind_data_schema.components.devices.Device(
                name="Lick sensor",
                model="1007079-1",
                manufacturer=aind_data_schema_models.organizations.Organization.TE_CONNECTIVITY,
            ),
            lick_sensor_type=aind_data_schema_models.devices.LickSensorType.PIEZOELECTIC,
        ),
    ],
)

SYNC_DAQ = aind_data_schema.components.devices.DAQDevice(
    name="Sync DAQ",
    manufacturer=aind_data_schema_models.organizations.Organization.NATIONAL_INSTRUMENTS,
    model="NI-6612",
    data_interface=aind_data_schema_models.devices.DataInterface.PCIE,
)
TASKCONTROL_DAQ = aind_data_schema.components.devices.DAQDevice(
    name="TaskControl DAQ",
    manufacturer=aind_data_schema_models.organizations.Organization.NATIONAL_INSTRUMENTS,
    model="6001",
    data_interface=aind_data_schema_models.devices.DataInterface.USB,
)
CAMSTIM_DAQ = aind_data_schema.components.devices.DAQDevice(
    name="Camstim DAQ",
    manufacturer=aind_data_schema_models.organizations.Organization.NATIONAL_INSTRUMENTS,
    model="6323",
    data_interface=aind_data_schema_models.devices.DataInterface.PCIE,
)
OPTO_DAQ = aind_data_schema.components.devices.DAQDevice(
    manufacturer=aind_data_schema_models.organizations.Organization.NATIONAL_INSTRUMENTS,
    name="Opto DAQ",
    model="9264",
    data_interface=aind_data_schema_models.devices.DataInterface.ETH,
)
EPHYS_DAQ = aind_data_schema.components.devices.DAQDevice(
    manufacturer=aind_data_schema_models.organizations.Organization.NATIONAL_INSTRUMENTS,
    name="Ephys DAQ",
    model="6133",
    data_interface=aind_data_schema_models.devices.DataInterface.PXI,
)

PHOTODIODE = aind_data_schema.components.devices.Detector(
    name="Stimulus photodiode",
    manufacturer=aind_data_schema_models.organizations.Organization.THORLABS,
    model="PDA25K",
    data_interface=aind_data_schema_models.devices.DataInterface.COAX,
    detector_type=aind_data_schema_models.devices.DetectorType.OTHER,
    cooling=aind_data_schema_models.devices.Cooling.NO_COOLING,
    notes="Photodiode used to measure stimulus monitor frame updates",
)
MICROPHONE = aind_data_schema.components.devices.Device(
    name="Stimulus microphone",
    manufacturer=aind_data_schema_models.organizations.Organization.DODOTRONIC,
    model="momimic",
    notes="Microphone used to record stimulus speaker output from approximately 50 mm",
)

LASER_488 = aind_data_schema.components.devices.Laser(
    name="Opto laser #1",
    manufacturer=aind_data_schema_models.organizations.Organization.VORTRAN,
    model="Stradus 488-50",
    wavelength=488,
    wavelength_unit=aind_data_schema_models.units.SizeUnit.NM,
)
LASER_633 = LASER_488.model_copy(
    update={
        "name": "Opto laser #2",
        "model": "Stradus 633-50",
        "wavelength": 633,
    }
)

LASER_GALVO_X = aind_data_schema.components.devices.AdditionalImagingDevice(
    name="Opto laser galvo X",
    manufacturer=aind_data_schema_models.organizations.Organization.THORLABS,
    model="GVS012",
    imaging_device_type=aind_data_schema.components.devices.ImagingDeviceType.GALVO,
)
LASER_GALVO_Y = LASER_GALVO_X.model_copy(
    update={"name": LASER_GALVO_X.name.replace("X", "Y")}
)

ALLIED_CAMERA = aind_data_schema.components.devices.Camera(
    name="Generic camera",
    manufacturer=aind_data_schema_models.organizations.Organization.ALLIED,
    model="G-032",
    data_interface=aind_data_schema_models.devices.DataInterface.ETH,
    cooling=aind_data_schema_models.devices.Cooling.NO_COOLING,
    frame_rate=60,
    frame_rate_unit=aind_data_schema_models.units.FrequencyUnit.HZ,
    sensor_width=636,
    sensor_height=508,
    size_unit=aind_data_schema_models.units.SizeUnit.PX,
    chroma=aind_data_schema_models.devices.CameraChroma.BW,
    recording_software=aind_data_schema.components.identifiers.Software(
        name="MultiVideoRecorder",
        version="1.1.7",
    ),
)
FRONT_CAMERA = ALLIED_CAMERA.model_copy(update={"name": "Front camera"})
FRONT_CAMERA_ASSEMBLY = aind_data_schema.components.devices.CameraAssembly(
    name="Front camera assembly",
    target=aind_data_schema.components.devices.CameraTarget.FACE,
    camera=FRONT_CAMERA,
    lens=aind_data_schema.components.devices.Lens(
        name="Front camera lens",
        manufacturer=aind_data_schema_models.organizations.Organization.EDMUND_OPTICS,
        model="86604",
        additional_settings=dict(
            focal_length=8.5,
            focal_length_unit=aind_data_schema_models.units.SizeUnit.MM,
        ),
    ),
    filter=aind_data_schema.components.devices.Filter(
        name="Front camera filter",
        manufacturer=aind_data_schema_models.organizations.Organization.SEMROCK,
        model="FF01-715",
        filter_type=aind_data_schema_models.devices.FilterType.LONGPASS,
        cut_on_wavelength=715,
        wavelength_unit=aind_data_schema_models.units.SizeUnit.NM,
    ),
    coordinate_system=CAMERA_COORDINATE_SYSTEM,
    relative_position=[
        aind_data_schema_models.coordinates.AnatomicalRelative.ANTERIOR,
        aind_data_schema_models.coordinates.AnatomicalRelative.SUPERIOR,
    ],
    transform=[
        aind_data_schema.components.coordinates.Affine(
            affine_transform=[
                [-0.17365, 0.98481, 0],
                [0.44709, 0.07883, -0.89101],
                [-0.87747, -0.15472, -0.45399],
                [0.154, 0.03078, 0.06346],
            ],
        ),
    ],
)

SIDE_CAMERA = ALLIED_CAMERA.model_copy(update={"name": "Side camera"})
SIDE_CAMERA_ASSEMBLY = aind_data_schema.components.devices.CameraAssembly(
    name="Side camera assembly",
    target=aind_data_schema.components.devices.CameraTarget.BODY,
    camera=SIDE_CAMERA,
    filter=aind_data_schema.components.devices.Filter(
        name="Side camera filter",
        manufacturer=aind_data_schema_models.organizations.Organization.SEMROCK,
        model="FF01-747",
        filter_type=aind_data_schema_models.devices.FilterType.BANDPASS,
        cut_on_wavelength=730,
        cut_off_wavelength=764,
        wavelength_unit=aind_data_schema_models.units.SizeUnit.NM,
    ),
    lens=aind_data_schema.components.devices.Lens(
        name="Side camera lens",
        manufacturer=aind_data_schema_models.organizations.Organization.NAVITAR,
        additional_settings=dict(
            focal_length=6.0,
            focal_length_unit=aind_data_schema_models.units.SizeUnit.MM,
        ),
    ),
    coordinate_system=CAMERA_COORDINATE_SYSTEM,
    relative_position=[
        aind_data_schema_models.coordinates.AnatomicalRelative.LEFT,
    ],
    transform=[
        aind_data_schema.components.coordinates.Affine(
            affine_transform=[
                [-1, 0, 0],
                [0, -1, 0],
                [-1, 0, -0.03617],
                [0.23887, -0.02535],
            ],
        ),
    ],
)

EYE_CAMERA = ALLIED_CAMERA.model_copy(
    update={
        "name": "Eye camera",
        "notes": "There is a mirror in the light path between the eye and the camera.",
    }
)
EYE_CAMERA_ASSEMBLY = aind_data_schema.components.devices.CameraAssembly(
    name="Eye camera assembly",
    target=aind_data_schema.components.devices.CameraTarget.EYE,
    camera=EYE_CAMERA,
    filter=aind_data_schema.components.devices.Filter(
        name="Eye camera filter",
        manufacturer=aind_data_schema_models.organizations.Organization.SEMROCK,
        model="FF01-850",
        filter_type=aind_data_schema_models.devices.FilterType.BANDPASS,
        cut_on_wavelength=845,
        cut_off_wavelength=855,
        wavelength_unit=aind_data_schema_models.units.SizeUnit.NM,
    ),
    lens=aind_data_schema.components.devices.Lens(
        name="Eye camera lens",
        manufacturer=aind_data_schema_models.organizations.Organization.INFINITY_PHOTO_OPTICAL,
        model="213073",
        additional_settings=dict(
            focal_length=6.0,
            focal_length_unit=aind_data_schema_models.units.SizeUnit.MM,
        ),
    ),
    coordinate_system=CAMERA_COORDINATE_SYSTEM,
    relative_position=[
        aind_data_schema_models.coordinates.AnatomicalRelative.RIGHT,
    ],
    transform=[
        aind_data_schema.components.coordinates.Affine(
            affine_transform=[
                [-0.5, -0.86603, 0],
                [-0.366, 0.21131, -0.90631],
                [0.78489, -0.45315, -0.42262],
                [-0.14259, 0.06209, 0.09576],
                [0.78489, -0.45315, -0.42262],
                [-0.14259, 0.06209, 0.09576],
            ],
        ),
    ],
)

NOSE_CAMERA = ALLIED_CAMERA.model_copy(update={"name": "Nose camera"})
NOSE_CAMERA_ASSEMBLY = aind_data_schema.components.devices.CameraAssembly(
    name="Nose camera assembly",
    target=aind_data_schema.components.devices.CameraTarget.FACE,
    camera=NOSE_CAMERA,
    filter=aind_data_schema.components.devices.Filter(
        name="Nose camera filter",
        manufacturer=aind_data_schema_models.organizations.Organization.SEMROCK,
        model="FF01-715",
        filter_type=aind_data_schema_models.devices.FilterType.LONGPASS,
        cut_on_wavelength=715,
        wavelength_unit=aind_data_schema_models.units.SizeUnit.NM,
    ),
    lens=aind_data_schema.components.devices.Lens(
        name="Nose camera lens",
        manufacturer=aind_data_schema_models.organizations.Organization.EDMUND_OPTICS,
        model="85360",
        additional_settings=dict(
            focal_length=25,
            focal_length_unit=aind_data_schema_models.units.SizeUnit.MM,
        ),
    ),
    relative_position=[
        aind_data_schema_models.coordinates.AnatomicalRelative.ANTERIOR,
        aind_data_schema_models.coordinates.AnatomicalRelative.LEFT,
    ],
    # !pending:
    # coordinate_system=CAMERA_COORDINATE_SYSTEM,
    # transform=[
    #     aind_data_schema.components.coordinates.Affine(
    #         affine_transform=[[], [], [], []]
    #     ),
    # ],
)
FRONT_LED = aind_data_schema.components.devices.LightEmittingDiode(
    manufacturer=aind_data_schema_models.organizations.Organization.AMS_OSRAM,
    name="Front camera illumination LED",
    model="LZ4-40R308-0000",
    wavelength=740,
    wavelength_unit=aind_data_schema_models.units.SizeUnit.NM,
)
SIDE_LED = FRONT_LED.model_copy(
    update={"name": FRONT_LED.name.replace("Front", "Side")}
)
NOSE_LED = FRONT_LED.model_copy(
    update={"name": FRONT_LED.name.replace("Front", "Nose")}
)

EYE_LED = aind_data_schema.components.devices.LightEmittingDiode(
    manufacturer=aind_data_schema_models.organizations.Organization.AMS_OSRAM,
    name="Eye camera illumination LED",
    model="LZ4-40R608-0000",
    wavelength=850,
    wavelength_unit=aind_data_schema_models.units.SizeUnit.NM,
)


def get_basestation(
    slot: Literal[2, 3], session: DynamicRoutingSession | None = None
) -> aind_data_schema.components.devices.NeuropixelsBasestation:
    probe_letters = "ABC" if slot == 2 else "DEF"
    basestation_firmware_version = "240.1120"
    bsc_firmware_version = "1.0.144"
    if session is not None and session.is_ephys:
        with contextlib.suppress(
            ValueError, FileNotFoundError, AttributeError, IndexError
        ):
            settings_xml_path = session.ephys_settings_xml_path
            text = settings_xml_path.read_text()
            basestation_firmware_version = re.search(r'bs_firmware_version="([\d.]+)"', text).group(1)  # type: ignore [union-attr]
            bsc_firmware_version = re.search(r'bsc_firmware_version="([\d.]+)"', text).group(1)  # type: ignore [union-attr]
    return aind_data_schema.components.devices.NeuropixelsBasestation(
        name=f"probes {probe_letters}",
        manufacturer=aind_data_schema_models.organizations.Organization.IMEC,
        model="PXIe",
        basestation_firmware_version=basestation_firmware_version,
        bsc_firmware_version=bsc_firmware_version,
        slot=slot,
        ports=[
            aind_data_schema.components.devices.ProbePort(
                index=i, probes=[f"probe{letter}"]
            )
            for i, letter in zip(range(3), probe_letters, strict=False)
        ],
    )


def get_computers(
    session: DynamicRoutingSession,
) -> list[aind_data_schema.components.devices.Computer]:
    rig_name = _normalize_rig_name(session.rig)
    computer_names = []
    if rig_name.startswith("og") or rig_name.startswith("np"):
        computer_names.extend(["STIM", "SYNC", "MON"])
    else:
        computer_names.append("BEH")
    if rig_name.startswith("np"):
        computer_names.append("ACQ")
    return [
        aind_data_schema.components.devices.Computer(
            name=name, operating_system="Windows 10"
        )
        for name in computer_names
    ]


def get_components_and_connections(
    session: DynamicRoutingSession,
) -> tuple[
    list[aind_data_schema.components.devices.Device],
    list[aind_data_schema.components.connections.Connection],
]:
    components = [
        DISC,
        MONITOR,
        SPEAKER,
        LICK_SPOUT_ASSEMBLY,
        CAMSTIM_DAQ,
        TASKCONTROL_DAQ,
        *get_computers(session),
    ]
    connections = [
        aind_data_schema.components.connections.Connection(
            source_device=LICK_SPOUT_ASSEMBLY.name,
            target_device=CAMSTIM_DAQ.name,
            send_and_receive=True,
        ),
        aind_data_schema.components.connections.Connection(
            source_device=DISC.name,
            target_device=CAMSTIM_DAQ.name,
        ),
        aind_data_schema.components.connections.Connection(
            target_device=TASKCONTROL_DAQ.name,
            source_device=SPEAKER.name,
        ),
    ]
    if is_behavior_box(session.rig):
        components.extend(
            [
                SIDE_CAMERA_ASSEMBLY,
                SIDE_LED,
            ]
        )
    if is_og_rig(session.rig) or is_np_rig(session.rig):
        components.extend(
            [
                SYNC_DAQ,
                MICROPHONE,
                PHOTODIODE,
                FRONT_CAMERA_ASSEMBLY,
                FRONT_LED,
                SIDE_CAMERA_ASSEMBLY,
                SIDE_LED,
                EYE_CAMERA_ASSEMBLY,
                EYE_LED,
                NOSE_CAMERA_ASSEMBLY,
                NOSE_LED,
                OPTO_DAQ,
                LASER_488,
                LASER_633,
                LASER_GALVO_X,
                LASER_GALVO_Y,
            ]
        )
        connections.extend(
            [
                aind_data_schema.components.connections.Connection(
                    source_device=source_device.name,
                    target_device=SYNC_DAQ.name,
                )
                for source_device in (
                    PHOTODIODE,
                    FRONT_CAMERA_ASSEMBLY.camera,
                    SIDE_CAMERA_ASSEMBLY.camera,
                    EYE_CAMERA_ASSEMBLY.camera,
                    NOSE_CAMERA_ASSEMBLY.camera,
                    LICK_SPOUT_ASSEMBLY.lick_spouts[0],
                    LASER_488,
                    LASER_633,
                )
            ]
        )
        connections.extend(
            [
                aind_data_schema.components.connections.Connection(
                    source_device=OPTO_DAQ.name,
                    target_device=target_device.name,
                )
                for target_device in (
                    LASER_488,
                    LASER_633,
                    LASER_GALVO_X,
                    LASER_GALVO_Y,
                )
            ]
        )
    if is_np_rig(session.rig):
        components.extend(
            [
                *get_ephys_assemblies(session),
                get_basestation(slot=2, session=session),
                get_basestation(slot=3, session=session),
                EPHYS_DAQ,
            ]
        )
        connections.extend(
            [
                aind_data_schema.components.connections.Connection(
                    source_device=source_device.name,
                    target_device=EPHYS_DAQ.name,
                )
                for source_device in (
                    SPEAKER,
                    MICROPHONE,
                    PHOTODIODE,
                )
            ]
        )
    return components, connections


def get_modalities(
    session: DynamicRoutingSession,
) -> list[aind_data_schema_models.modalities.Modality]:
    modality = aind_data_schema_models.modalities.Modality
    modalities = [modality.BEHAVIOR, modality.BEHAVIOR_VIDEOS]
    if is_np_rig(session.rig):
        modalities.append(modality.ECEPHYS)
    return modalities


def get_ephys_assemblies(
    session: DynamicRoutingSession,
) -> list[aind_data_schema.components.devices.EphysAssembly]:
    """Get ephys assemblies for the session."""
    if not session.is_ephys:
        return []
    assemblies = []
    for probe_letter in ["A", "B", "C", "D", "E", "F"]:
        assemblies.append(
            aind_data_schema.components.devices.EphysAssembly(
                name=probe_letter,
                manipulator=aind_data_schema.components.devices.Manipulator(
                    name=f"Manipulator {probe_letter}",
                    manufacturer=aind_data_schema_models.organizations.Organization.NEW_SCALE_TECHNOLOGIES,
                    model="M3-LS-3.4-15",
                ),
                probes=[
                    aind_data_schema.components.devices.EphysProbe(
                        name=f"probe{probe_letter}",
                        manufacturer=aind_data_schema_models.organizations.Organization.IMEC,
                        probe_model=aind_data_schema.components.devices.ProbeModel.NP1,
                    )
                ],
            )
        )
    return assemblies


def get_instrument_model(
    session: DynamicRoutingSession,
) -> aind_data_schema.core.instrument.Instrument:
    """Get the Pydantic model corresponding to the 'instrument.json' for a given session."""

    return aind_data_schema.core.instrument.Instrument(
        location=get_location(session.rig),
        instrument_id=session.rig,
        modification_date=session.session_start_time.date(),
        modalities=get_modalities(session),
        calibrations=[],
        coordinate_system=RIG_COORDINATE_SYSTEM,
        temperature_control=None,
        notes=None,
        connections=(
            components_and_connections := get_components_and_connections(session)
        )[1],
        components=components_and_connections[0],
    )


if __name__ == "__main__":
    session = DynamicRoutingSession("814666_20251107")
    metadata = get_instrument_model(session)
    print(metadata.model_dump_json(indent=2))
    with open("instrument_814666_2025-11-07.json", "w") as f:
        f.write(metadata.model_dump_json(indent=2))



"""
Create pydantic metadata models for the current variants of sessions (DR, Templeton, surface
recordings).

Code is tightly coupled to current version of npc_sessions, so it doesn't make sense to split into a
different repo, but required dependencies are made optional.
"""

try:
    import aind_data_schema  # noqa
except ImportError:
    raise ImportError(
        "Optional dependencies are required to use this module: install `npc_sessions[metadata]`"
    ) from None

from npc_sessions.aind_data_schema.acquisition import get_acquisition_model
from npc_sessions.aind_data_schema.data_description import get_data_description_model
from npc_sessions.aind_data_schema.instrument import get_instrument_model

__all__ = [
    "get_acquisition_model",
    "get_data_description_model",
    "get_instrument_model",
]
